import 'dart:math';
import 'package:grizzly/grizzly.dart';
import 'package:collection/collection.dart';
import 'package:text_table/text_table.dart';

extension MatIterableExt<T> on Iterable<List<T>> {
  bool isEqual(Iterable<Iterable<T>> v) {
    final iter1 = iterator;
    final iter2 = v.iterator;
    while (iter1.moveNext() && iter2.moveNext()) {
      if (!iter1.current.isEqual(iter2.current)) {
        return false;
      }
    }
    return true;
  }

  bool isEqualVector(Iterable<T> v) {
    for (final row in this) {
      if (!row.isEqual(v)) return false;
    }
    return true;
  }

  bool isEqualScalar(T v) {
    for (final row in this) {
      if (!row.isEqualScalar(v)) return false;
    }
    return true;
  }

  /*
  void addColumn(Iterable<T> col) {
    final thisIter = iterator;
    final colIter = col.iterator;
    while (thisIter.moveNext() && colIter.moveNext()) {
      thisIter.current.add(colIter.current);
    }
  }

  set assign(Iterable<Iterable<T>> value) {
    final thisIter = iterator;
    final valIter = value.iterator;
    while (thisIter.moveNext() && valIter.moveNext()) {
      thisIter.current.assign = valIter.current;
    }
  }
   */
}

extension GenericIterableIterableExt<T> on Iterable<Iterable<T>> {
  int get numRows => length;

  int get numCols => isEmpty ? 0 : first.length;

  Iterable<T> operator [](int index) => elementAt(index);

  Iterable<T> array() sync* {
    for (final list in this) {
      yield* list;
    }
  }

  Iterable<Iterable<T>> get cols => ColsView(this);

  Iterable<T> unique() => array().unique();

  Iterable<T> get diagonal => mapIndexed((i, e) => e.elementAt(i));

  ColsView<T> get transpose => ColsView(this);

  List<List<T>> toList2D() => map((e) => e.toList()).toList();

  String tableString({TableRenderer renderer = const TableRenderer()}) {
    final list = <List>[];

    for (final row in this) {
      list.add(row.map((e) => e?.toString() ?? '').toList());
    }

    return renderer.render(list);
  }
}

extension MatExt<T> on List<List<T>> {
  int get numRows => length;

  int get numCols => isEmpty ? 0 : first.length;

  Index2D get shape => Index2D(numRows, numCols);

  int get numCells => fold(0, (current, element) => current + element.length);

  Cols<T> get cols => Cols(this);

  void apply(T Function(T v) applier) {
    for (final row in this) {
      row.apply(applier);
    }
  }

  bool get isSquare => numRows == numCols;

  // TODO reshape

  List<List<T>> get transpose {
    final ret = List<List<T>>.generate(numCols, (index) => <T>[]);
    for (int i = 0; i < numRows; i++) {
      for (int j = 0; j < numCols; j++) {
        ret[j].add(this[i][j]);
      }
    }
    return ret;
  }

  List<T> get diagonal {
    final ret = <T>[];
    for (int i = 0; i < min(numRows, numCols); i++) {
      ret.add(this[i][i]);
    }
    return ret;
  }

  set diagonal(List<T> other) {
    for (int i = 0; i < [numRows, numCols, other.length].min; i++) {
      this[i][i] = other[i];
    }
  }

  set diagonalSame(T other) {
    for (int i = 0; i < min(numRows, numCols); i++) {
      this[i][i] = other;
    }
  }

  List<List<S>> shaped<S>(S v) =>
      List<List<S>>.generate(numRows, (index) => List<S>.filled(numCols, v));

  List<List<T>> clone() => List<List<T>>.generate(
      numRows, (r) => List<T>.generate(numCols, (c) => this[r][c]));

  List<List<T>> cut(Index2D start, [Index2D? end]) {
    final Index2D myShape = shape;
    if (end == null) {
      end = myShape;
    } else {
      if (end < Index2D.zero) {
        throw ArgumentError.value(end, 'end', 'Index out of range!');
      }
      if (end >= myShape) {
        throw ArgumentError.value(end, 'end', 'Index out of range!');
      }
      if (start > end) {
        throw ArgumentError.value(end, 'end', 'Must be greater than start!');
      }
    }
    if (start < Index2D.zero) {
      throw ArgumentError.value(start, 'start', 'Index out of range!');
    }
    if (start >= myShape) {
      throw ArgumentError.value(start, 'start', 'Index out of range!');
    }

    final list = <List<T>>[];

    for (int r = start.row; r < end.row; r++) {
      list.add(this[r].cut(start.col, end.col));
    }

    return list;
  }
}

abstract class MatView {
  static Iterable<Iterable<T>> generate<T>(
          int rows, int cols, T Function(int i, int j) generator) =>
      Iterable.generate(
          rows, (i) => Iterable<T>.generate(cols, (j) => generator(i, j)));

  static Iterable<Iterable<T>> filled<T>(int rows, int cols, T fill) =>
      Iterable.generate(rows, (i) => ConstantIterable(fill, cols));

  static Iterable<Iterable<T>> identity<T>(int size, T diagonal, T fill) =>
      Iterable.generate(size,
          (i) => Iterable.generate(size, (j) => i == j ? diagonal : fill));

  static Iterable<Iterable<T>> diagonal<T>(Iterable<T> diagonal, T fill) =>
      diagonal.diagonal(fill);

  static List<List<T>> fromColumn<T>(Iterable<T> column) =>
      fromColumns([column]);

  static List<List<T>> fromColumns<T>(Iterable<Iterable<T>> cols) {
    final iterators = cols.map((e) => e.iterator).toList();

    final ret = <List<T>>[];
    while (true) {
      if (!iterators.every((element) => element.moveNext())) {
        break;
      }
      ret.add(iterators.map((e) => e.current).toList());
    }

    return ret;
  }

  static List<List<E>> sized<E>(int rows, int cols, E fill) =>
      List<List<E>>.generate(rows, (index) => List<E>.filled(cols, fill));
}

abstract class Mat {
  static List<List<T>> generate<T>(
          int rows, int cols, T Function(int i, int j) generator) =>
      List.generate(
          rows, (i) => List<T>.generate(cols, (j) => generator(i, j)));

  static List<List<double>> identity(int size, {double fill = 1.0}) =>
      List.generate(size, (i) => List.generate(size, (j) => i == j ? fill : 0));

  static List<List<int>> intIdentity(int size, {int fill = 1}) =>
      List.generate(size, (i) => List.generate(size, (j) => i == j ? fill : 0));

  static List<List<T>> diagonal<T>(Iterable<T> diagonal, T fill) =>
      List.generate(
          diagonal.length,
          (i) => List.generate(
              diagonal.length, (j) => i == j ? diagonal.elementAt(i) : fill));

  static List<List<int>> intDiagonal(Iterable<int> values, {int fill = 0}) =>
      diagonal(values, fill);

  static List<List<double>> doubleDiagonal(Iterable<double> values,
          {double fill = 0.0}) =>
      diagonal(values, fill);

  static List<List<num>> numDiagonal(Iterable<num> values, {num fill = 0}) =>
      diagonal(values, fill);

  static List<List<T>> fromColumn<T>(Iterable<T> column) =>
      fromColumns([column]);

  static List<List<T>> fromColumns<T>(Iterable<Iterable<T>> cols) {
    final iterators = cols.map((e) => e.iterator).toList();

    final ret = <List<T>>[];
    while (true) {
      if (!iterators.every((element) => element.moveNext())) {
        break;
      }
      ret.add(iterators.map((e) => e.current).toList());
    }

    return ret;
  }

  static List<List<E>> sized<E>(int rows, int cols, E fill) =>
      List<List<E>>.generate(rows, (index) => List<E>.filled(cols, fill));
}
