import 'dart:collection';

import 'package:grizzly/grizzly.dart';
import 'package:text_table/text_table.dart';

class NamedIterable<T> extends IterableBase<T> {
  String name;

  final Iterable<T> _inner;

  NamedIterable(this._inner, {this.name = ''});

  @override
  Iterator<T> get iterator => _inner.iterator;

  @override
  String toString() => (name.isNotEmpty ? '$name: ' : '') + _inner.toString();
}

class NamedList<T> extends ListBase<T> implements List<T>, NamedIterable<T> {
  @override
  String name;

  @override
  final List<T> _inner;

  NamedList(this._inner, {this.name = ''});

  @override
  set length(int newLength) => _inner.length = newLength;

  @override
  int get length => _inner.length;

  @override
  void operator []=(int index, T value) => _inner[index] = value;

  @override
  T operator [](int index) => _inner[index];

  @override
  String toString() => (name.isNotEmpty ? '$name: ' : '') + _inner.toString();
}

class Named2DView<T> extends IterableBase<List<T>> {
  final List<String> names;

  final Iterable<List<T>> _inner;

  Named2DView(this._inner, {List<String>? names})
      : names = names?.toList() ??
            List<String>.generate(_inner.numCols, (i) => 'col$i') {
    if (numCols != this.names.length) {
      throw Exception('number of columns does not match number of names');
    }
  }

  @override
  Iterator<List<T>> get iterator => _inner.iterator;

  List<T> operator [](int index) => _inner[index];

  int loc(String name) => names.indexOf(name);

  @override
  String toString({TableRenderer renderer = const TableRenderer()}) {
    final list = <List>[];

    for (final row in this) {
      list.add(row.map((e) => e?.toString() ?? '').toList());
    }

    return renderer.render(list, columns: names);
  }

  Grouped<T> group(List<Grouper> groupers) => Grouped<T>(this, groupers);

  Map<String, T> rowMap(int r) =>
      indices(numCols).map((c) => MapEntry(names[c], this[r][c])).toMap();

  Iterable<Map<String, T>> get rowMaps => indices(numRows).map(rowMap);
}

class Named2D<T> extends ListBase<List<T>>
    implements List<List<T>>, Named2DView<T> {
  @override
  final List<String> names;

  @override
  final List<List<T>> _inner;

  Named2D(this._inner, {List<String>? names}) : names = names?.toList() ?? [];

  @override
  set length(int newLength) => _inner.length = newLength;

  @override
  int get length => _inner.length;

  @override
  void operator []=(int index, List<T> value) => _inner[index] = value;

  @override
  List<T> operator [](int index) => _inner[index];

  @override
  int loc(String name) => names.indexOf(name);

  @override
  String toString({TableRenderer renderer = const TableRenderer()}) {
    final list = <List>[];

    for (final row in this) {
      list.add(row.map((e) => e?.toString() ?? '').toList());
    }

    return renderer.render(list, columns: names);
  }

  void addMap(Map<String, dynamic> row) {
    _inner.add(names.map((e) => row[e]).cast<T>().toList());
  }

  void addMaps(List<Map<String, dynamic>> rows) {
    for (final row in rows) {
      addMap(row);
    }
  }

  @override
  Grouped<T> group(List<Grouper> groupers) => Grouped<T>(this, groupers);

  @override
  Map<String, T> rowMap(int r) =>
      indices(numCols).map((c) => MapEntry(names[c], this[r][c])).toMap();

  @override
  Iterable<Map<String, T>> get rowMaps => indices(numRows).map(rowMap);

  /*
  @override
  String toString(
      {int? maxRows = 100, TableRenderer renderer = const TableRenderer()}) {
    int only = maxRows != null && length > maxRows ? maxRows : length;
    final list = <List>[];
    for (int i = 0; i < only; i++) {
      list.add(this[i].values.toList());
    }
    final output = renderer.render(list, columns: names);
    return output +
        (only != length ? '\n...showing only $only rows out of $length' : '');
  }
   */
}

extension Named2DListExt<T> on List<List<T>> {
  Named2D<T> named(List<String> names) => Named2D(this, names: names);
}

extension Named2DMapExt<T> on Iterable<Map<String, T>> {
  Named2D<T> named({List<String>? names}) {
    if (isEmpty) {
      return <List<T>>[].named([]);
    }

    names = names ?? first.keys.toList();
    final rows = <List<T>>[];
    for (final row in this) {
      rows.add(names.map((key) => row[key]).cast<T>().toList());
    }
    return rows.named(names);
  }
}

extension MapEntryIterableExt<K, V> on Iterable<MapEntry<K, V>> {
  Map<K, V> toMap() => Map<K, V>.fromEntries(this);
}
