import 'dart:math' as math;

import 'package:grizzly/src/utils/comap.dart';
import 'package:grizzly/src/utils/norepeat_sampler.dart';
import 'package:grizzly/src/utils/reducers.dart';
import 'package:text_table/text_table.dart';
import 'package:grizzly/grizzly.dart';

extension NumExt on num {
  int get sign => isNegative ? -1 : (this == 0 ? 0 : 1);

  double get sqrt => math.sqrt(this);
}

typedef Num1DView = Iterable<num>;

extension NullableNumIterableExt<T extends num> on Iterable<T?> {
  T? get min {
    T? ret;
    for (final item in this) {
      if (item == null) continue;
      if (ret == null || item < ret) {
        ret = item;
      }
    }
    return ret;
  }

  T? get max {
    T? ret;
    for (final item in this) {
      if (item == null) continue;
      if (ret == null || item > ret) {
        ret = item;
      }
    }
    return ret;
  }

  Extent<T>? get extent {
    T? min;
    T? max;
    for (final item in this) {
      if (item == null) continue;
      if (max == null || item > max) max = item;
      if (min == null || item < min) min = item;
    }
    if (min == null || max == null) {
      return null;
    }
    return Extent<T>(min, max);
  }

  T? get ptp {
    final range = extent;
    if (range == null) {
      return null;
    }
    return range.distance;
  }

  T get mode {
    throw UnimplementedError();
  }

  T? get median {
    if (length == 0) return null;
    final list = toList()..zort();
    return list[length ~/ 2];
  }

  double get sum {
    double ret = 0;
    for (final item in this) {
      if (item == null) continue;
      ret += item;
    }
    return ret;
  }

  double sumMask(Iterable<bool> mask) {
    double ret = 0;
    for (final item in this) {
      if (item == null) continue;
      ret += item;
    }
    return ret;
  }

  double get prod {
    double ret = 1;
    for (final item in this) {
      if (item == null) continue;
      ret *= item;
    }
    return ret;
  }

  double get mean {
    if (length == 0) return 0.0;

    final reducer = MeanReducer();
    for(var item in this) {
      reducer.consume(item ?? 0);
    }
    return reducer.mean;
  }

  double average(Iterable<num?> weights) {
    if (length == 0) return 0.0;

    if (weights.length != length) {
      throw LengthMismatch(
          expected: length, found: weights.length, subject: 'weights');
    }

    double sum = 0.0;
    num denom = 0.0;

    final iter = iterator;
    for (final weight in weights) {
      final item = (iter..moveNext()).current;
      if (item == null || weight == null) continue;
      sum += item * weight;
      denom += weight;
    }
    return sum / denom;
  }

  double get variance {
    if (length == 0) return 0.0;

    final reducer = VarianceReducer();
    for(var item in this) {
      reducer.consume(item ?? 0);
    }
    return reducer.variance;
  }

  double get std => math.sqrt(variance);

  int countIf(bool Function(T? item) condition) {
    int ret = 0;
    for (final item in this) {
      if (condition(item)) ret++;
    }
    return ret;
  }

  double cov(Iterable<T?> y) {
    if (y.length != length) {
      throw LengthMismatch(expected: length, found: length, subject: 'y');
    }
    if (length == 0) return 0.0;

    final double meanX = mean;
    final double meanY = y.mean;

    final reducer = MeanReducer();
    for(final item in zip(this, y)) {
      if (item.a == null || item.b == null) continue;
      reducer.consume((item.a! - meanX) * (item.b! - meanY));
    }

    return reducer.mean;
  }

/* TODO covMatrix
  Numeric1D<double> covMatrix(Numeric2D y) {
    if (y.numRows != length) throw Exception('Size mismatch!');
    final double meanX = mean;
    final Double1D meanY = y.col.mean;
    Double1D sum = Double1D.sized(y.numCols);
    for (int i = 0; i < length; i++) {
      sum += (y.col[i] - meanY) * (values.elementAt(i) - meanX);
    }
    return sum / length;
  }
 */

  double corrcoef(Iterable<T?> y) {
    if (y.length != length) {
      throw LengthMismatch(expected: length, found: length, subject: 'y');
    }
    return cov(y) / (std * y.std);
  }

  /* TODO corrcoefMatrix
    Numeric1D<double> corrcoefMatrix(Numeric2D y) {
    if (y.numRows != length) throw Exception('Size mismatch!');
    return covMatrix(y) / (y.std * std);
  }
   */

  String describe({TableRenderer renderer = const TableRenderer()}) {
    return renderer.render([
      ['Count (non-null)', countNonNull],
      ['Mean', mean],
      ['std', std],
      ['min', min],
      ['max', max],
      ['P2P', ptp],
      // TODO
    ], columns: [
      'Stat',
      'Value'
    ]);
  }

  num dot([Iterable<num?>? other]) => cofold<num, num?, num?>(0, this,
      other ?? this, (value, num? a, num? b) => value + (a ?? 0) * (b ?? 0));

  Iterable<num> matmult(List<List<num>> matrix) =>
      matrix.cols.map((col) => dot(col));

  Iterable<num?> operator -() sync* {
    for (final v in this) {
      if (v == null) yield null;
      yield -v!;
    }
  }

  Iterable<num> operator +(/* Iterable<num> | num */ other) {
    if (other is num) {
      return map((e) => (e ?? 0) + other);
    } else if (other is Iterable<num>) {
      return comap<num, num?, num?>(this, other, (a, b) => (a ?? 0) + (b ?? 0));
    }

    throw UnsupportedError('only num and Iterable<num> are supported');
  }

  Iterable<num> operator -(/* Iterable<num> | num */ other) {
    if (other is num) {
      return map((e) => (e ?? 0) - other);
    } else if (other is Iterable<num>) {
      return comap<num, num?, num?>(this, other, (a, b) => (a ?? 0) - (b ?? 0));
    }

    throw UnsupportedError('only num and Iterable<num> are supported');
  }

  Iterable<num> operator *(/* Iterable<num> | num */ other) {
    if (other is num) {
      return map((e) => (e ?? 0) * other);
    } else if (other is Iterable<num>) {
      return comap<num, num?, num?>(this, other, (a, b) => (a ?? 0) * (b ?? 0));
    }

    throw UnsupportedError('only num and Iterable<num> are supported');
  }

  Iterable<double> operator /(/* Iterable<num> | num */ other) {
    if (other is num) {
      return map((e) => (e ?? 0) / other);
    } else if (other is Iterable<num>) {
      return comap<double, num?, num?>(
          this, other, (a, b) => (a ?? 0) / (b ?? 0));
    }

    throw UnsupportedError('only num and Iterable<num> are supported');
  }

  Iterable<int> operator ~/(/* Iterable<num> | num */ other) {
    if (other is num) {
      return map((e) => (e ?? 0) ~/ other);
    } else if (other is Iterable<num>) {
      return comap<int, num?, num?>(
          this, other, (a, b) => (a ?? 0) ~/ (b ?? 0));
    }

    throw UnsupportedError('only num and Iterable<num> are supported');
  }
}

extension NumIterableExt<T extends num> on Iterable<T> {
  bool isEqual(Iterable<num> v, {double eps = 1e-8}) {
    if (length != v.length) return false;
    final iter1 = iterator;
    final iter2 = v.iterator;
    while (iter1.moveNext() && iter2.moveNext()) {
      if ((iter1.current - iter2.current).abs() > eps) {
        return false;
      }
    }
    return true;
  }

  void assertEqual(Iterable<num> v, {double eps = 1e-8}) {
    if (length != v.length) {
      throw AssertionError('Length mismatch');
    }
    final iter1 = iterator;
    final iter2 = v.iterator;
    int i = 0;
    while (iter1.moveNext() && iter2.moveNext()) {
      final diff = (iter1.current - iter2.current).abs();
      if (diff > eps) {
        throw AssertionError('$i =>  ${iter1.current} != ${iter2.current}');
      }
      i++;
    }
  }

  bool isEqualScalar(num v, {double eps = 1e-8}) {
    for (final item in this) {
      if ((item - v).abs() > eps) return false;
    }
    return true;
  }

  T get min {
    if (isEmpty) {
      throw Exception('min called on empty list');
    }

    T ret = first;
    for (final item in this) {
      if (item < ret) {
        ret = item;
      }
    }
    return ret;
  }

  T get max {
    if (isEmpty) {
      throw Exception('max called on empty list');
    }

    T ret = first;
    for (final item in this) {
      if (item > ret) {
        ret = item;
      }
    }
    return ret;
  }

  Extent<T> get extent => Extent(min, max);

  Iterable<double> sqrt() => map((e) => math.sqrt(e));

  Iterable<num> abs() => map((e) => e.abs());

  Iterable<num> operator -() sync* {
    for (final v in this) {
      yield -v;
    }
  }

  /// Returns matrix multiplication c = [this] * [other]. Interprets this as matrix
  /// of form n x 1. Interprets the [other] as form 1 x m. The returned result is
  /// of the form n x m.
  List<List<num>> matmultRowVector(Iterable<num> other) {
    final ret = <List<num>>[];
    for (final x in this) {
      final row = <num>[];
      for (final y in other) {
        row.add(x * y);
      }
      ret.add(row);
    }
    return ret;
  }

  Iterable<double> get log => map((e) => math.log(e));

  Iterable<double> get log10 => map((e) => math.log(e) / math.ln10);

  Iterable<double> logN(double n) => map((e) => math.log(e) / math.log(n));

  Iterable<double> get exp => map((e) => math.exp(e));

  Iterable<num> pow(num p) => map((e) => math.pow(e, p));

  Iterable<double> get floorToDouble => map((e) => e.floorToDouble());

  Iterable<double> get ceilToDouble => map((e) => e.ceilToDouble());

  Iterable<int> get floor => map((e) => e.floor());

  Iterable<int> get ceil => map((e) => e.ceil());

  List<double> toDouble() => map((e) => e.toDouble()).toList();

  Iterable<double> doubleView() => map((e) => e.toDouble());

  List<int> toInt() => map((e) => e.toInt()).toList();

  Iterable<int> intView() => map((e) => e.toInt());

  Iterable<double> sin() => map((e) => math.sin(e));

  Iterable<double> cos() => map((e) => math.cos(e));

  Iterable<double> tan() => map((e) => math.tan(e));

  Iterable<double> acos() => map((e) => math.acos(e));

  Iterable<double> asin() => map((e) => math.asin(e));

  Iterable<double> atan() => map((e) => math.atan(e));
}
